#include "precomp.h"
#include "split3pm.h"

BITRATES KnownBitrates[] = {
{  32,  32,  32,  32,  32,   8,   8 },
{  32,  32,  32,  32,  32,   8,   8 },
{  64,  48,  40,  64,  48,  16,  16 },
{  96,  56,  48,  96,  56,  24,  24 },
{ 128,  64,  56, 128,  64,  32,  32 },
{ 160,  80,  64, 160,  80,  64,  64 },
{ 192,  96,  80, 192,  96,  80,  80 },
{ 224, 112,  96, 224, 112,  56,  96 },
{ 256, 128, 112, 256, 128,  64, 112 },
{ 288, 160, 128, 288, 160, 128, 128 },
{ 320, 192, 160, 320, 192, 160, 160 },
{ 352, 224, 192, 352, 224, 112, 192 },
{ 384, 256, 224, 384, 256, 128, 224 },
{ 416, 320, 256, 416, 320, 256, 256 },
{ 448, 384, 320, 448, 384, 320, 320 },
{ -1, -1, -1, -1, -1, -1 }
}; 

DWORD ExtractBits( DWORD dwSource, DWORD dwFrom, DWORD dwTo )
{
    DWORD dwResult = 0;
    DWORD dwResultBitIndex = 1;

    for( DWORD dwSourceBitIndex = dwFrom; dwSourceBitIndex <= dwTo; dwSourceBitIndex++ )
    {
        if( dwSource & (1UL << dwSourceBitIndex) )
            dwResult |= dwResultBitIndex;
        dwResultBitIndex <<= 1;
    }
    return dwResult;
}

void FillMp3Head( LPMP3HEAD h, DWORD dwSource )
{
    printf( "MP3 Header: 0x%08lx\n", dwSource );
    h->dwEmphasis = ExtractBits(dwSource,0,1);
    h->dwOriginal = ExtractBits(dwSource,2,2);
    h->dwCopyright = ExtractBits(dwSource,3,3);
    h->dwModeExtension = ExtractBits(dwSource,4,5);
    h->dwChannelMode = ExtractBits(dwSource,6,7);
    h->dwPrivateBit = ExtractBits(dwSource,8,8);
    h->dwPaddingBit = ExtractBits(dwSource,9,9);
    h->dwSampleRateFreq = ExtractBits(dwSource,10,11);

    h->dwBitrateIndex = ExtractBits(dwSource,12,15);

    h->dwFrameSync = ExtractBits(dwSource,21,31);
    h->dwAudioVersion = ExtractBits(dwSource,19,20);
    h->dwLayerDescription = ExtractBits(dwSource,17,18);
    h->dwProtection = ExtractBits(dwSource,16,16);

    DWORD dwIndex;

//    bits V1,L1 V1,L2 V1,L3 V2,L1 V2,L2 V2,L3 

    //         0  1     2     3     4     5     6
    //    bits V1,L1 V1,L2 V1,L3 V2,L1 V2,L2 V2,L3 
    if( h->dwAudioVersion == MPEG_Version_1 )
    {
        if( h->dwLayerDescription == MPEG_Layer_1 )
            dwIndex = 1;
        else if( h->dwLayerDescription == MPEG_Layer_2 )
            dwIndex = 2;
        else if( h->dwLayerDescription == MPEG_Layer_3 )
            dwIndex = 6;
        else
            dwIndex = 0;
    }
    else if( h->dwAudioVersion == MPEG_Version_2 )
    {
        if( h->dwLayerDescription == MPEG_Layer_3 )     
            dwIndex = 3;
        else if( h->dwLayerDescription == MPEG_Layer_1 )
            dwIndex = 4;
        else 
            dwIndex = 5;
    }
    else dwIndex = 6;

    h->dwBitrate = KnownBitrates[h->dwBitrateIndex].x[dwIndex] * 1000;

//    F 2 (11,10) Sampling rate frequency index (values are in Hz) bits MPEG1 MPEG2 MPEG2.5 
//    00 44100 22050 11025 
//    01 48000 24000 12000 
//    10 32000 16000 8000 
//    11 reserv. reserv. reserv. 
    
    if( h->dwSampleRateFreq == 0 )
    {
        if( h->dwAudioVersion == MPEG_Version_1 )
            h->dwSampleRate = 44100;
        else if( h->dwAudioVersion == MPEG_Version_2 )
            h->dwSampleRate = 22050;
        else
            h->dwSampleRate = 11025;
    }
    else if( h->dwSampleRateFreq == 1 )
    {
        if( h->dwAudioVersion == MPEG_Version_1 )
            h->dwSampleRate = 48000;
        else if( h->dwAudioVersion == MPEG_Version_2 )
            h->dwSampleRate = 24000;
        else
            h->dwSampleRate = 12000;
    }
    else
    {
        if( h->dwAudioVersion == MPEG_Version_1 )
            h->dwSampleRate = 32000;
        else if( h->dwAudioVersion == MPEG_Version_2 )
            h->dwSampleRate = 16000;
        else
            h->dwSampleRate = 8000;
    }
}          

#ifdef TARGET_IS_UNIX
static int dwIgnore;
#else
static DWORD dwIgnore;
#endif
BOOL bInfoOnly = FALSE;
BOOL bQuiet = FALSE;


BOOL MP3File::CreateTarget(LPCSTR lpszFileName, DWORD dwFrom, DWORD dwTo,BOOL bAppend)
{
    BOOL bSuccess = FALSE;

    if( dwTo < dwFrom )
    {
        DWORD dwTemp = dwTo;
        dwTo = dwFrom;
        dwFrom = dwTo;
    }
    printf( "About to write %s from byte %ld to byte %ld\n", lpszFileName, dwFrom, dwTo );

    BOOL bCanAccessFile = FALSE;

#ifdef TARGET_IS_UNIX
    int hTargetFile = creat( lpszFileName, S_IRUSR|S_IRGRP|S_IROTH|S_IWUSR|S_IWGRP|S_IWOTH );
    if( hTargetFile >= 0 )
        bCanAccessFile = TRUE;
    else printf( "ERROR %ld, Unable to open %s for writing\n", GetLastError(), lpszFileName );
#else
    HANDLE hTargetFile;
    if( bAppend )
    {
        hTargetFile = CreateFile( lpszFileName, GENERIC_WRITE|GENERIC_READ, FILE_SHARE_READ, NULL, CREATE_NEW|OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0 );
        if( hTargetFile != INVALID_HANDLE_VALUE )
        {
            bCanAccessFile = TRUE;
            SetFilePointer(hTargetFile,0,0,FILE_END);
        }
        else printf( "ERROR %ld, Unable to open %s for write-append\n", GetLastError(), lpszFileName );
    }
    else
    {
        hTargetFile = CreateFile( lpszFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0 );
        if( hTargetFile != INVALID_HANDLE_VALUE )
            bCanAccessFile = TRUE;
        else printf( "ERROR %ld, Unable to open %s for writing\n", GetLastError(), lpszFileName );
    }
#endif

    if( bCanAccessFile )
    {
#ifdef TARGET_IS_UNIX
        if( lseek(m_hFile,(LONG)dwFrom,SEEK_SET) < 0 )
#else
        if( SetFilePointer(m_hFile,(LONG)dwFrom,0,FILE_BEGIN) == INVALID_SET_FILE_POINTER )
#endif
            printf("ERROR %ld, SetFilePointer() failed\n", GetLastError(), dwFrom );
        else
        {
            DWORD dwTotalBytes = dwTo - dwFrom;
            bSuccess = TRUE;
            while( dwTotalBytes )
            {
                DWORD dwBytesToCopy = dwTotalBytes;
                if( dwBytesToCopy > m_dwCacheMemorySize )
                    dwBytesToCopy = m_dwCacheMemorySize;

#ifdef TARGET_IS_UNIX
                if( read( m_hFile, m_lpbCache, dwBytesToCopy) < 0 )
#else
                if( !ReadFile(m_hFile,m_lpbCache,dwBytesToCopy,&dwIgnore,0) )
#endif
                {
                    printf("ERROR %ld, unable to read from source file\n", GetLastError() );
                    bSuccess = FALSE;
                    break;
                }
#ifdef TARGET_IS_UNIX
                if( write(hTargetFile,m_lpbCache,dwBytesToCopy) < 0 )
#else
                if( !WriteFile(hTargetFile,m_lpbCache,dwBytesToCopy,&dwIgnore,0) )
#endif
                {
                    printf("ERROR %ld, Unable to write to target file\n", GetLastError() );
                    bSuccess = FALSE;
                    break;
                }
                dwTotalBytes -= dwBytesToCopy;
            }
        }
#ifdef TARGET_IS_UNIX
        close(hTargetFile);
#else
        CloseHandle(hTargetFile);
#endif
    }    
    return bSuccess;
}

BOOL MP3File::LocatePosition(DWORD dwSeconds, LPDWORD lpdwPosition)
{
    DWORD dwEstimatedPos = dwSeconds * m_dwBytesPerSecond;
    if( dwEstimatedPos > m_dwFileSize )
    {
        printf( "ERROR, position %d seconds out of file range (0..%d)\n", dwSeconds, m_dwLengthInSeconds );
        return FALSE;
    }
    DWORD dwStartReadPos = ( dwEstimatedPos > m_dwBytesPerFrame ) ? (dwEstimatedPos-m_dwBytesPerFrame+2) : 0;

#ifdef TARGET_IS_UNIX
	int nResult = lseek(m_hFile,(LONG)dwStartReadPos,SEEK_SET);
	if( nResult < 0 )
#else
    if( SetFilePointer(m_hFile,(LONG)dwStartReadPos,0,FILE_BEGIN) == INVALID_SET_FILE_POINTER )
#endif
    {
        printf( "ERROR %ld, SetFilePointer(%ld) failed\n", GetLastError(), dwStartReadPos );
        return FALSE;
    }

#ifdef TARGET_IS_UNIX
    dwIgnore = (DWORD) read( m_hFile, m_bBuffer, sizeof(m_bBuffer) );
	if( dwIgnore < 0 )
#else
    if( !ReadFile(m_hFile,m_bBuffer,sizeof(m_bBuffer),&dwIgnore,0) )
#endif
    {
        printf("ERROR %ld, ReadFile(%d bytes) failed at position\n", GetLastError(), sizeof(m_bBuffer), dwStartReadPos );
        return FALSE;
    }
    LPBYTE lpbPos = FindNextFrameStart(m_bBuffer,m_bBuffer+dwIgnore);
    if( !lpbPos )
    {
        printf( "ERROR, no frame found arround file position %ld\n", dwStartReadPos );
		return FALSE;
    }
    *lpdwPosition = dwStartReadPos + (lpbPos-m_bBuffer);

    printf( "Position for %ld:%ld is %ld\n", dwSeconds/60,dwSeconds%60, *lpdwPosition);
    return TRUE;
}

void MP3File::DumpInfo()
{
    printf( "File Size In Bytes: %ld\n", m_dwFileSize );
    printf( "File Size In Frames: %ld\n", m_dwNumberOfFrames );
    printf( "Bytes Per Frame: %ld\n", m_dwBytesPerFrame );
    printf( "Bytes Per Second: %ld\n", m_dwBytesPerSecond );

    DWORD dwMinutes = m_dwLengthInSeconds / 60;
    DWORD dwHours = 0;
    if( dwMinutes > 60 )
    {
        dwHours = dwMinutes / 60;
        dwMinutes = dwMinutes % 60;
    }

    printf( "Estimated Length: %ld:%ld:%ld (%ld seconds)\n", 
		dwHours, 
		dwMinutes, 		
		(DWORD) (m_dwLengthInSeconds % 60), (long) m_dwLengthInSeconds );
#define DUMPBIT(x) // printf( #x " = 0x%08lx\n", m_Header.dw##x );
#define DUMPDEC(x) printf( #x " = %ld\n", m_Header.dw##x );

    DUMPBIT(FrameSync)
    DUMPBIT(AudioVersion);
    DUMPBIT(LayerDescription);
    DUMPBIT(Protection);
    DUMPBIT(BitrateIndex);
    DUMPDEC(Bitrate);
    DUMPBIT(SampleRateFreq);
    DUMPDEC(SampleRate);
    DUMPBIT(PaddingBit);
    DUMPBIT(PrivateBit);
    DUMPBIT(ChannelMode);
    DUMPBIT(ModeExtension);
    DUMPBIT(Copyright);
    DUMPBIT(Original);
    DUMPBIT(Emphasis);
} 

LPBYTE MP3File::FindNextFrameStart(LPBYTE lpbBuffer,LPBYTE lpbMax)
{
    while(lpbBuffer < lpbMax)
    {
        if( *lpbBuffer == 0xFF )
        {
            if( lpbBuffer[1] == 0xFB )
            {
                return lpbBuffer;
            }
        }
        lpbBuffer++;
    }
    return NULL;
}


MP3File::MP3File()
{
#ifdef TARGET_IS_UNIX
    m_hFile = -1;
#else
    m_hFile = INVALID_HANDLE_VALUE;
#endif
    m_lpbCache = 0;
}

BOOL MP3File::Initialize()
{
    m_dwCacheMemorySize = 1024 * 1024 * 10;
    if( !m_lpbCache )
        m_lpbCache = new BYTE[m_dwCacheMemorySize];

    return (m_lpbCache == NULL) ? FALSE : TRUE;
}

MP3File::~MP3File()
{
#ifdef TARGET_IS_UNIX
    if( m_hFile >= 0 )
        close(m_hFile);
#else
    if( m_hFile != INVALID_HANDLE_VALUE )
    {
        CloseHandle(m_hFile);
    }
#endif
    if( m_lpbCache )
    {
        delete m_lpbCache;
    }
}

#ifdef TARGET_IS_UNIX
DWORD GetFileSize(int m_hFile, void* )
{
    long pos = lseek(m_hFile,0,SEEK_CUR);
    long size = lseek(m_hFile,0,SEEK_END);
    lseek(m_hFile,pos,SEEK_SET);
    return (DWORD) size;
}
#endif

BOOL MP3File::LoadFile( LPCSTR lpszFileName )
{
    BOOL bSuccess = FALSE;

    printf( "Loading %s...\n", lpszFileName );

#ifdef TARGET_IS_UNIX
    if( m_hFile >= 0 )
        close(m_hFile);
		
	m_hFile = open(lpszFileName,O_RDONLY);
	if( m_hFile >= 0 )
#else
    if( m_hFile != INVALID_HANDLE_VALUE )
        CloseHandle(m_hFile);
    m_hFile = CreateFile( lpszFileName, GENERIC_READ, FILE_SHARE_READ, NULL,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0 );
    if( m_hFile != INVALID_HANDLE_VALUE )
#endif
    {
        m_dwFileSize = GetFileSize(m_hFile,0);

        if( (m_dwFileSize > MAX_BUFFER) && (m_dwFileSize != (DWORD)-1) )
        {
#ifdef TARGET_IS_UNIX
            if( read( m_hFile, m_bBuffer, sizeof(m_bBuffer)) > 0 )
#else
            if( ReadFile(m_hFile,m_bBuffer,sizeof(m_bBuffer),&dwIgnore,0) )
#endif
            {
                if( (m_bBuffer[0] == 0xFF) && (m_bBuffer[1] == 0xFB) )
                {
                    DWORD dwMask = 0;

                    dwMask |= ((DWORD)m_bBuffer[3]);
                    dwMask |= ((DWORD)m_bBuffer[2]) << 8;
                    dwMask |= ((DWORD)m_bBuffer[1]) << 16;
                    dwMask |= ((DWORD)m_bBuffer[0]) << 24;

                    FillMp3Head( &m_Header, dwMask );

                    m_dwBytesPerFrame = 144 * m_Header.dwBitrate/ m_Header.dwSampleRate;

                    m_dwNumberOfFrames = m_dwFileSize / m_dwBytesPerFrame;
                    double temp = 38.4, secs;
                    secs = (double)m_dwNumberOfFrames;  
                    secs /= temp;
                    temp *= (double) m_dwBytesPerFrame;
                    m_dwBytesPerSecond = (DWORD)temp;
					m_dwLengthInSeconds = m_dwFileSize;
                    m_dwLengthInSeconds /= m_dwBytesPerSecond;
                    bSuccess = TRUE;
                }
                else printf( "ERROR, %s is not a valid MP3 file.\n"
                             "The first two bytes read %02hx%02hx (expected FFFB)\n",
                              lpszFileName, (USHORT)m_bBuffer[0],(USHORT)m_bBuffer[1]);
            }
            else printf( "ERROR %ld, ReadFile(%s) failed\n", GetLastError(), lpszFileName );
        }
        else printf( "ERROR %ld, %s is not a valid MP3 file (size = %d!)\n", GetLastError(), lpszFileName, m_dwFileSize );

        if( !bSuccess )
        {
#ifdef TARGET_IS_UNIX
            close(m_hFile);
            m_hFile = -1;
#else
            CloseHandle(m_hFile);
            m_hFile = INVALID_HANDLE_VALUE;
#endif
        }
    }
    else printf( "ERROR %ld, Unable to open file %s\n", GetLastError(), lpszFileName );

    return bSuccess;
} 

static bool DECINT(int& result,LPCSTR& p)
{
    if( (*p >= '0') && (*p <= '9') )
    {
        result = *(p++) - '0'; 
        return true;
    }
    return false;
}

static bool GETINT(int& result,LPCSTR& p)
{
    if( DECINT(result,p) )
    {
        int temporary;
        if( DECINT(temporary,p) )
        {
            result *= 10;
            result += temporary;
        }
        return true;
    }
    return false;
}

bool ParseTime(LPCSTR& pszExpression,LPDWORD lpdwTime)
{
    int c = 0, a = 0, b = 0;
    if( GETINT(a,pszExpression) )
        if( *(pszExpression++) == ':' )
            if( GETINT(b,pszExpression) )
                if( *pszExpression == ':' )
                {
                    pszExpression++;
                    if( GETINT(c,pszExpression) )
                    {
                        *lpdwTime = b*60 + a*60*60 + c;
                        return true;
                    }
                }
                else 
                {
                    *lpdwTime = b + a*60;
                    return true;
                }
    return false;
}

s3pPart::s3pPart( DWORD dwFrom, DWORD dwTo, LPCSTR lpszFileName )
    :   m_dwFrom(dwFrom), m_dwTo(dwTo), m_strFileName(lpszFileName)
{
}

BOOL s3pFile::Read( LPCSTR lpszFileName )
{
#ifdef TARGET_IS_UNIX
    FILE* fp = fopen( lpszFileName, "r" );
#else
    FILE* fp = fopen( lpszFileName, "rt" );
#endif
    if( fp != NULL )
    {
        CHAR szBuffer[1024];
        while( !feof( fp ) )
        {
            if( fgets( szBuffer, sizeof(szBuffer), fp ) )
            {
                LPSTR p = &(szBuffer[strlen(szBuffer)-1]);
                if( *p == '\n' )
                    *p = 0;
                
                if( !OnParse( szBuffer ) )
                    break;
            }
        }
        fclose( fp );
        return TRUE;													    
    }
    else printf( "ERROR %ld, fopen(%s) failed", GetLastError(), lpszFileName );
    return FALSE;
}

BOOL s3pFile::OnParse( LPCSTR lpszLine )
{
    LPSTR p = strchr(lpszLine,';');
    if( p )
        *p = 0;

    if( !strnicmp(lpszLine,"Source=",7) )
    {
        m_strSource = lpszLine+7;
    }
    else if( *lpszLine )
    {
        DWORD dwFrom, dwTo;
        LPCSTR p = lpszLine;

        if( ParseTime(p,&dwFrom) )
            if( *(p++) == '-' )
                if( ParseTime(p,&dwTo) )
                    if( *(p++) == ',' )
                    {
                        m_Parts.AddTail( new s3pPart(dwFrom,dwTo,p) );
                        return TRUE;
                    }
        
        printf( "Warning, ignoring line \"%s\"\n", lpszLine );
    }
    return TRUE;
}

int help()
{
    printf( "USAGE: split3pm [OPTIONS] {script}]\n\n"
        "OPTIONS:\n"
        "/S<File> ..... Source MP3 file\n"
        "/T<File> ..... Target MP3 file\n"
        "/F<Time> ..... Timespec.\n" 
        "/I ........... Display info only.\n" );  
    
    printf("; this is a sample Split3PM script\n"
#ifdef TARGET_IS_UNIX
        "Source=/mp3/filename.mp3\n"
        "\n"
        "; parts to extract:\n"
        "10:00-12:00,/mp3/test/f0.mp3\n"
        "14:00-16:00,/mp3/test/f1.mp3\n"
        "18:00-20:00,/mp3/test/f2.mp3\n" 
#else
        "Source=d:\\mp3\\filename.mp3\n"
        "\n"
        "; parts to extract:\n"
        "10:00-12:00,d:\\f0.mp3\n"
        "14:00-16:00,d:\\f1.mp3\n"
        "18:00-20:00,d:\\f2.mp3\n" 
#endif
        "\n"
        "For more information, please go to http://www.p-nand-q.com/split3pm.htm\n" );

    return 10;
}

s3pPart* GetCommandLineTimeSpec( LPCSTR lpszTimeSpec, LPCSTR lpszFilename )
{
    DWORD dwFrom, dwTo;

    if( ParseTime(lpszTimeSpec,&dwFrom) )
        if( *(lpszTimeSpec++) == '-' )
            if( ParseTime(lpszTimeSpec,&dwTo) )
                return new s3pPart(dwFrom,dwTo,lpszFilename);

    return NULL;
}

int main( int argc, char* argv[] )
{
#ifdef TARGET_IS_UNIX
    printf( "SPLIT3PM for Linux - Version 1.2 - Freeware\n"
#else
    printf( "SPLIT3PM for Win32 - Version 1.2 - Freeware\n"
#endif
            "Written by Gerson Kurz (Gerson.Kurz@t-online.de)\n\n" );
    
    PList Scripts; 

    LPCSTR lpszSource = NULL;
    LPCSTR lpszTarget = NULL;
    LPCSTR lpszTimeSpec = NULL;

    for( int i = 1; i < argc; i++ )
    {
        char* arg = argv[i];
        if( (*arg == '-') 
#ifndef TARGET_IS_UNIX
            || (*arg == '/') 
#endif
            )
        {
            switch(toupper(arg[1]))
            {
            case 'S':
                lpszSource = arg+2;
                break;
            case 'T':
                lpszTarget = arg+2;
                break;
            case 'F':
                lpszTimeSpec = arg+2;
                break;
            case 'I':
                bInfoOnly = TRUE;
                break;
            case 'Q':
                bQuiet = TRUE;
                break;
            default:
                return help();
            }
        }
        else Scripts.AddTail( new PString(arg) );
    }

    if( lpszSource && ((lpszTarget && lpszTimeSpec) || bInfoOnly) )
    {
        s3pPart* pPart = GetCommandLineTimeSpec( lpszTimeSpec, lpszTarget );
        if( pPart )
        {
            MP3File file;
            if( file.Initialize() )
                if( file.LoadFile(lpszSource) )
                {
                    DWORD dwFrom, dwTo;

                    file.DumpInfo();
                    if( !bInfoOnly )
                        if( file.LocatePosition(pPart->m_dwFrom, &dwFrom) )
                            if( file.LocatePosition(pPart->m_dwTo, &dwTo) )
                                file.CreateTarget(pPart->m_strFileName,dwFrom,dwTo,FALSE);
                }

            delete pPart;
        }
        else printf( "ERROR, invalid timespec \"%s\"\n", lpszTimeSpec );
    }
    else if( lpszSource || lpszTarget || lpszTimeSpec )
        return help();
    
    if( !Scripts.m_lCount && !lpszSource )
        return help();

    ENUMERATE( &Scripts, PString, pS )
    {
        LPCSTR lpszPath = *pS;
        LPCSTR lpszExtension = strrchr(lpszPath,'.');

        if( lpszExtension && !stricmp(lpszExtension,".mp3") )
        {
            MP3File file;
            if( file.Initialize() )
                if( file.LoadFile(lpszPath) )
                    file.DumpInfo();
        }
        else
        {
            printf( "About to process script %s\n", lpszPath );
            s3pFile f;
            if( f.Read( lpszPath ) )
            {
                if( IsEmptyString(f.m_strSource) )
                    printf( "ERROR, missing Source= in input script\n" );
                else
                {
                    MP3File file;
                    if( file.Initialize() )
                    {
                        if( file.LoadFile(f.m_strSource) )
                        {
                            DWORD dwFrom, dwTo;

                            file.DumpInfo();

                            if( !bInfoOnly )
                            {
                                PString strLastKnownFilename;
                                ENUMERATE( &f.m_Parts, s3pPart, pPart )
                                {
                                    if( file.LocatePosition(pPart->m_dwFrom, &dwFrom) )
                                    {
                                        if( file.LocatePosition(pPart->m_dwTo, &dwTo) )
                                        {
                                            LPCSTR p = pPart->m_strFileName;
                                            if( !strcmp(p,"+") )
                                            {
                                                if( IsEmptyString(strLastKnownFilename) )
                                                    break;
                                                if( !file.CreateTarget(strLastKnownFilename,dwFrom,dwTo,TRUE) )
                                                    break;
                                            }
                                            else
                                            {
                                                if( !file.CreateTarget(pPart->m_strFileName,dwFrom,dwTo,FALSE) )
                                                    break;
                                                strLastKnownFilename = pPart->m_strFileName;
                                            }
                                        }
                                        else break;
                                    }
                                    else break;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
	return 0;
}



